-- see https://markkarpov.com/tutorial/megaparsec.html
{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE OverloadedStrings #-}

module Main.Parser.Megaparsec (parseExpr, ParseResult) where

import Control.Monad.Combinators.Expr
import Data.Functor.Identity qualified
import Data.Text.Lazy
import Data.Void (Void)
import Main.Types.Expr
import Text.Megaparsec as MP
import Text.Megaparsec.Char as C
import Text.Megaparsec.Char.Lexer as L

type Parser = Parsec Void Text

lexemeParser :: Parser a -> Parser a
lexemeParser = L.lexeme C.space

symbolParser :: Text -> Parser Text
symbolParser = L.symbol C.space

intParser :: Parser Int
intParser = lexemeParser L.decimal

term :: Parser Expr
term = Lit <$> intParser

table :: [[Operator Parser Expr]]
table =
  [ [ binaryOp "*" Mul,
      binaryOp "/" Div
    ],
    [ binaryOp "+" Add,
      binaryOp "-" Sub
      -- binaryOp "." Pop
    ]
  ]

binaryOp :: Text -> (a -> a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
binaryOp name f = InfixL (f <$ symbolParser name)

prefixOp :: Text -> (a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
prefixOp name f = Prefix (f <$ symbolParser name)

postfixOp :: Text -> (a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
postfixOp name f = Postfix (f <$ symbolParser name)

expr :: Parser Expr
expr = makeExprParser term table

type ParseResult = Either (ParseErrorBundle Text Void) Expr

parseExpr :: Text -> ParseResult
parseExpr = MP.parse (C.space *> expr <* eof) ""

-- parseE = parseExpr